@@ -1,3 +1,6 @@
+use ct_codecs::{Base64UrlSafeNoPadding, Encoder};
+use hmac_sha1_compact::Hash as SHA1;
+use hmac_sha256::Hash as SHA256;
 use hmac_sha512::sha384 as hmac_sha384;
 use rsa::pkcs1::{FromRsaPrivateKey as _, FromRsaPublicKey as _};
 use rsa::pkcs8::{FromPrivateKey as _, FromPublicKey as _, ToPrivateKey as _, ToPublicKey as _};
@@ -66,11 +69,14 @@ impl RSAPublicKey {
 
 #[doc(hidden)]
 #[derive(Debug, Clone)]
-pub struct RSAKeyPair(rsa::RsaPrivateKey);
+pub struct RSAKeyPair {
+    rsa_sk: rsa::RsaPrivateKey,
+    metadata: Option<KeyMetadata>,
+}
 
 impl AsRef<rsa::RsaPrivateKey> for RSAKeyPair {
     fn as_ref(&self) -> &rsa::RsaPrivateKey {
-        &self.0
+        &self.rsa_sk
     }
 }
 
@@ -80,7 +86,10 @@ impl RSAKeyPair {
             .or_else(|_| rsa::RsaPrivateKey::from_pkcs1_der(der))?;
         rsa_sk.validate()?;
         rsa_sk.precompute()?;
-        Ok(RSAKeyPair(rsa_sk))
+        Ok(RSAKeyPair {
+            rsa_sk,
+            metadata: None,
+        })
     }
 
     pub fn from_pem(pem: &str) -> Result<Self, Error> {
@@ -89,25 +98,28 @@ impl RSAKeyPair {
             .or_else(|_| rsa::RsaPrivateKey::from_pkcs1_pem(pem))?;
         rsa_sk.validate()?;
         rsa_sk.precompute()?;
-        Ok(RSAKeyPair(rsa_sk))
+        Ok(RSAKeyPair {
+            rsa_sk,
+            metadata: None,
+        })
     }
 
     pub fn to_der(&self) -> Result<Vec<u8>, Error> {
-        self.0
+        self.rsa_sk
             .to_pkcs8_der()
             .map_err(Into::into)
             .map(|x| x.as_ref().to_vec())
     }
 
     pub fn to_pem(&self) -> Result<String, Error> {
-        self.0
+        self.rsa_sk
             .to_pkcs8_pem()
             .map_err(Into::into)
             .map(|x| x.to_string())
     }
 
     pub fn public_key(&self) -> RSAPublicKey {
-        let rsa_pk = self.0.to_public_key();
+        let rsa_pk = self.rsa_sk.to_public_key();
         RSAPublicKey(rsa_pk)
     }
 
@@ -118,7 +130,10 @@ impl RSAKeyPair {
         };
         let mut rng = rand::thread_rng();
         let rsa_sk = rsa::RsaPrivateKey::new(&mut rng, modulus_bits)?;
-        Ok(RSAKeyPair(rsa_sk))
+        Ok(RSAKeyPair {
+            rsa_sk,
+            metadata: None,
+        })
     }
 }
 
@@ -126,6 +141,8 @@ pub trait RSAKeyPairLike {
     fn jwt_alg_name() -> &'static str;
     fn key_pair(&self) -> &RSAKeyPair;
     fn key_id(&self) -> &Option<String>;
+    fn metadata(&self) -> &Option<KeyMetadata>;
+    fn attach_metadata(&mut self, metadata: KeyMetadata) -> Result<(), Error>;
     fn hash(message: &[u8]) -> Vec<u8>;
     fn padding_scheme(&self) -> rsa::PaddingScheme;
 
@@ -133,11 +150,8 @@ pub trait RSAKeyPairLike {
         &self,
         claims: JWTClaims<CustomClaims>,
     ) -> Result<String, Error> {
-        let jwt_header = JWTHeader {
-            algorithm: Self::jwt_alg_name().to_string(),
-            key_id: self.key_id().clone(),
-            ..Default::default()
-        };
+        let jwt_header = JWTHeader::new(Self::jwt_alg_name().to_string(), self.key_id().clone())
+            .with_metadata(self.metadata());
         Token::build(&jwt_header, claims, |authenticated| {
             let digest = Self::hash(authenticated.as_bytes());
             let mut rng = rand::thread_rng();
@@ -204,6 +218,15 @@ impl RSAKeyPairLike for RS256KeyPair {
         &self.key_id
     }
 
+    fn metadata(&self) -> &Option<KeyMetadata> {
+        &self.key_pair.metadata
+    }
+
+    fn attach_metadata(&mut self, metadata: KeyMetadata) -> Result<(), Error> {
+        self.key_pair.metadata = Some(metadata);
+        Ok(())
+    }
+
     fn hash(message: &[u8]) -> Vec<u8> {
         hmac_sha256::Hash::hash(message).to_vec()
     }
@@ -320,6 +343,14 @@ impl RS256PublicKey {
         self.key_id = Some(key_id.to_string());
         self
     }
+
+    pub fn sha1_thumbprint(&self) -> String {
+        Base64UrlSafeNoPadding::encode_to_string(SHA1::hash(&self.pk.to_der().unwrap())).unwrap()
+    }
+
+    pub fn sha256_thumbprint(&self) -> String {
+        Base64UrlSafeNoPadding::encode_to_string(SHA256::hash(&self.pk.to_der().unwrap())).unwrap()
+    }
 }
 
 //
@@ -349,6 +380,15 @@ impl RSAKeyPairLike for RS512KeyPair {
         &self.key_id
     }
 
+    fn metadata(&self) -> &Option<KeyMetadata> {
+        &self.key_pair.metadata
+    }
+
+    fn attach_metadata(&mut self, metadata: KeyMetadata) -> Result<(), Error> {
+        self.key_pair.metadata = Some(metadata);
+        Ok(())
+    }
+
     fn hash(message: &[u8]) -> Vec<u8> {
         hmac_sha512::Hash::hash(message).to_vec()
     }
@@ -465,6 +505,14 @@ impl RS512PublicKey {
         self.key_id = Some(key_id.to_string());
         self
     }
+
+    pub fn sha1_thumbprint(&self) -> String {
+        Base64UrlSafeNoPadding::encode_to_string(SHA1::hash(&self.pk.to_der().unwrap())).unwrap()
+    }
+
+    pub fn sha256_thumbprint(&self) -> String {
+        Base64UrlSafeNoPadding::encode_to_string(SHA256::hash(&self.pk.to_der().unwrap())).unwrap()
+    }
 }
 
 //
@@ -494,6 +542,15 @@ impl RSAKeyPairLike for RS384KeyPair {
         &self.key_id
     }
 
+    fn metadata(&self) -> &Option<KeyMetadata> {
+        &self.key_pair.metadata
+    }
+
+    fn attach_metadata(&mut self, metadata: KeyMetadata) -> Result<(), Error> {
+        self.key_pair.metadata = Some(metadata);
+        Ok(())
+    }
+
     fn hash(message: &[u8]) -> Vec<u8> {
         hmac_sha384::Hash::hash(message).to_vec()
     }
@@ -610,6 +667,14 @@ impl RS384PublicKey {
         self.key_id = Some(key_id.to_string());
         self
     }
+
+    pub fn sha1_thumbprint(&self) -> String {
+        Base64UrlSafeNoPadding::encode_to_string(SHA1::hash(&self.pk.to_der().unwrap())).unwrap()
+    }
+
+    pub fn sha256_thumbprint(&self) -> String {
+        Base64UrlSafeNoPadding::encode_to_string(SHA256::hash(&self.pk.to_der().unwrap())).unwrap()
+    }
 }
 
 //
@@ -639,6 +704,15 @@ impl RSAKeyPairLike for PS256KeyPair {
         &self.key_id
     }
 
+    fn metadata(&self) -> &Option<KeyMetadata> {
+        &self.key_pair.metadata
+    }
+
+    fn attach_metadata(&mut self, metadata: KeyMetadata) -> Result<(), Error> {
+        self.key_pair.metadata = Some(metadata);
+        Ok(())
+    }
+
     fn hash(message: &[u8]) -> Vec<u8> {
         hmac_sha256::Hash::hash(message).to_vec()
     }
@@ -784,6 +858,15 @@ impl RSAKeyPairLike for PS512KeyPair {
         &self.key_id
     }
 
+    fn metadata(&self) -> &Option<KeyMetadata> {
+        &self.key_pair.metadata
+    }
+
+    fn attach_metadata(&mut self, metadata: KeyMetadata) -> Result<(), Error> {
+        self.key_pair.metadata = Some(metadata);
+        Ok(())
+    }
+
     fn hash(message: &[u8]) -> Vec<u8> {
         hmac_sha512::Hash::hash(message).to_vec()
     }
@@ -900,6 +983,14 @@ impl PS512PublicKey {
         self.key_id = Some(key_id.to_string());
         self
     }
+
+    pub fn sha1_thumbprint(&self) -> String {
+        Base64UrlSafeNoPadding::encode_to_string(SHA1::hash(&self.pk.to_der().unwrap())).unwrap()
+    }
+
+    pub fn sha256_thumbprint(&self) -> String {
+        Base64UrlSafeNoPadding::encode_to_string(SHA256::hash(&self.pk.to_der().unwrap())).unwrap()
+    }
 }
 
 //
@@ -929,6 +1020,15 @@ impl RSAKeyPairLike for PS384KeyPair {
         &self.key_id
     }
 
+    fn metadata(&self) -> &Option<KeyMetadata> {
+        &self.key_pair.metadata
+    }
+
+    fn attach_metadata(&mut self, metadata: KeyMetadata) -> Result<(), Error> {
+        self.key_pair.metadata = Some(metadata);
+        Ok(())
+    }
+
     fn hash(message: &[u8]) -> Vec<u8> {
         hmac_sha384::Hash::hash(message).to_vec()
     }
@@ -1045,4 +1145,12 @@ impl PS384PublicKey {
         self.key_id = Some(key_id.to_string());
         self
     }
+
+    pub fn sha1_thumbprint(&self) -> String {
+        Base64UrlSafeNoPadding::encode_to_string(SHA1::hash(&self.pk.to_der().unwrap())).unwrap()
+    }
+
+    pub fn sha256_thumbprint(&self) -> String {
+        Base64UrlSafeNoPadding::encode_to_string(SHA256::hash(&self.pk.to_der().unwrap())).unwrap()
+    }
 }
